//+------------------------------------------------------------------+
//|                                                     moexlink.mqh |
//|                                    Copyright (c) 2025, Marketeer |
//|                           https://www.mql5.com/ru/articles/16964 |
//+------------------------------------------------------------------+
#include "toyjson2.mqh"
#include <MQL5Book/MqlError.mqh>

#define MOEX_LINK_LOG_NO      0
#define MOEX_LINK_LOG_YES     1
#define MOEX_LINK_LOG_DETAILS 2
#define MOEX_LINK_LOG_VERBOSE 3

class MoexLink
{
protected:
   int httpcode;
   string headers;
   string response;

   template<typename T>
   static bool contains(const T &a[], const T v)
   {
     for(int i = 0; i < ArraySize(a); i++)
     {
       if(a[i] == v) return true;
     }
     return false;
   }
   
   static string escape(const string s)
   {
     uchar a[];
     uchar forbidden[] = {'<', '>', '/', '\\', '"', ':', '|', '*', '?'};
     int n = StringToCharArray(s, a);
     for(int i = 0; i < n - 1; i++)
     {
       if(a[i] < 32 || contains(forbidden, a[i]))
       {
         a[i] = '_';
       }
     }
     return CharArrayToString(a);
   }
   
   static string mangle(const string filename, const string ext = ".txt")
   {
      const static string root = "MOEX";
      const static string today = "/today/";
      MqlDateTime mdt;
      TimeLocal(mdt);
      FileIsExist(root + today);
      // _LastError == 5018 -> subdirectory exists
      const string folder = (_LastError == 5018) ? today : StringFormat("/%04d%02d%02d/", mdt.year, mdt.mon, mdt.day);
      const string fullpath = root + folder + escape(filename) + ext;
      return fullpath;
   }
   
public:
   struct Settings
   {
      ushort logging; // logging level
      bool dumping;
      bool testing;
      uint timeout; // seconds
      uint throttle; // milliseconds
      string parameters; // common for all requests, such as "iss.meta=off"

      Settings(ushort loglevel = 0, bool dump = false, bool test = false, uint time = 5, uint wait = 0, string pstr = NULL)
      {
         logging = loglevel;
         dumping = dump;
         testing = test;
         timeout = time;
         throttle = wait;
         parameters = pstr;
      }

      Settings(const Settings &copy)
      {
         this = copy;
      }
   } settings;
   
   MoexLink() { }
   MoexLink(Settings &s): settings(s) { }
   
   void applySettings(Settings &s)
   {
      settings = s;
   }

   // customize headers for authentication and other purposes
   void setHeaders(const string h)
   {
      headers = h;
   }
   
   string getLastResponse() const
   {
      return response;
   }
   
   int getLastHttpCode() const
   {
      return httpcode;
   }

   string buildQuery(const string params) // key=value&...
   {
      string query = settings.parameters;
      if(StringLen(query) && StringLen(params))
      {
         query += "&";
      }
      
      query += params;
      if(StringLen(query)) query = "?" + query;
      return query;
   }

   string buildQuery(JsValue &params)
   {
      string query;
      string keys[];
      params.getProperties(keys);
      for(int i = 0; i < ArraySize(keys); i++)
      {
         if(StringLen(query))
         {
            query += "&";
         }
         
         query += keys[i] + (params[i].t != JS_NULL ? ("=" + params[i].s) : "");
      }
      // NB: urlencode is performed by WebRequest internally
      return buildQuery(query);
   }
   
   
   int fetch(const string url, char &result[])
   {
      if(settings.logging) Print("Get ", url);
      static ulong nexttime = 0;
      while(nexttime > GetTickCount64() && !IsStopped())
      {
         Sleep(settings.throttle / 2);
      }
      char body[];
      httpcode = WebRequest("GET", url, headers, settings.timeout * 1000, body, result, response);
      if(settings.logging) Print("Code: ", httpcode);
      if(settings.logging > 1) Print("Response: ", response);
      if(settings.throttle) nexttime = GetTickCount64() + settings.throttle;
      return httpcode;
   }

   string fetch(const string address)
   {
      char result[];
      if(settings.logging) PrintFormat("Fetching %s", address);
      const string storage = mangle(address);
      bool refresh = false;
      ResetLastError();
      if(settings.testing)
      {
         if(settings.logging) PrintFormat("Loading dump '%s'", storage);
         if(FileLoad(storage, result) > -1)
         {
            httpcode = 200; // can differ from actual, such as 404, 304, 5xx etc, stored in the dump
         }
         else
         {
            httpcode = 404;
            // seems useful to fall through for downloading if dumping is true
            if(settings.dumping)
            {
               fetch(address, result);
               refresh = true;
            }
         }
      }
      else
      {
         fetch(address, result);
      }
      if(!ArraySize(result))
      {
         if(settings.logging) PrintFormat("No data, error: %d, %s", _LastError, E2S(_LastError));
         return NULL;
      }

      if(settings.dumping && (!settings.testing || refresh))
      {
         const bool b = FileSave(storage, result);
         if(settings.logging) PrintFormat("File '%s' saved: %s", storage, (string)b);
      }
      
      const string text = CharArrayToString(result, 0, WHOLE_ARRAY, CP_UTF8);
      if(settings.logging > 2) Print("Data: ", text);
      
      return text;
   }
};
//+------------------------------------------------------------------+
